import time, random, textwrap

class State:
    def __init__(self, name, description):
        self.name = name
        self.description = description
        self.edges = {}
        self.objects = {}

class Edge:
    def __init__(self, name, from_node, commands, to_node, desc, health, constraintsname=[]):
        self.name = name
        self.from_node = from_node
        self.commands = commands
        self.to_node = to_node
        self.desc = desc
        self.health = health
        self.constraintsname = constraintsname
        self.constraint = {}

class Object:
    def __init__(self, location, name, shortdesc, description, weight=0.0):
        self.location = location #Where the object is
        self.name = name #Names of the object
        self.shortdesc = shortdesc #A short description
        self.description = description #The long description of the object
        self.weight = eval(weight) #The weight of the object
        
class Constraint:
    def __init__(self, name, objectname, state=None):
        self.name = name #Short unique name of the constraint
        self.objectname = objectname #What is needed for the constraint to be filled (such as an object)
        self.state = state #Where the object needs to be
        
class Backpack:
    def __init__(self, maxweight, value=0.0):
        self.maxweight = eval(maxweight)
        self.value = eval(value)
        self.objects = {}

class Engine:
    def __init__(self, filename):
        self.filename = filename
        self.initial_state = None
        self.backpack = None
        self.health = 10
        self.counter = 0
        self.initial_time = 0

    def load(self):
        self.initial_time = time.time()
        self.states = {}
        edges = []
        objects = []
        constraints = {}
        fp = open(self.filename)
        for line in fp:
            if len(line) == 0 or line[0] == '#':
                continue
            items = line.strip().split("::")
            if items[0] == 'node':
                s = State(*items[1:])
                if self.initial_state is None:
                    self.initial_state = s
                self.states[items[1]] = s
            elif items[0] == "backpack":
                b = Backpack(*items[1:])
                self.backpack = b
            elif items[0] == 'edge':
                if len(items) > 7:
                    cons = items[-1].strip().split(", ")
                    items[-1] = cons
                com = items[3].strip().split(", ")
                items[3] = com
                to = items[4].strip().split(", ")
                items[4] = to
                edges.append(Edge(*items[1:]))
            elif items[0] == "object":
                names = items[2].strip().split(", ")
                items[2] = names
                objects.append(Object(*items[1:]))
            elif items[0] == "constraint":
                c = Constraint(*items[1:])
                constraints[items[1]] = c
        for edge in edges:
            if len(edge.constraintsname) > 0:
                for c in edge.constraintsname:
                    edge.constraint[c] = constraints[c]
            self.states[edge.from_node].edges[edge.name] = edge
        for obj in objects:
            if obj.location == "backpack":
                self.backpack.objects[obj.name[0]] = obj
                self.backpack.value += obj.weight
            else:
                self.states[obj.location].objects[obj.name[0]] = obj

    def play(self):
        current_state = self.initial_state
        backpack = self.backpack
        self.time = time.time() - self.initial_time
        print(textwrap.fill('Welcome to the game. Type the direction you would like to go to move around, an action you would like to do to do it, "get [item name]" to pick up a specific item, "drop [item name]" to drop an item, "look [item name]" to more closely examine an item, "use [item name]" to use an item, "backpack" to see what you are carrying, "health" to see your health, and "time" to see how much time has elapsed. Type "help" if you are stuck to see a list of your options and directions you can go.'))
        print(" ")
        print("*********************************")
        while self.health > 0:
            self.time = time.time() - self.initial_time
            print(" ")
            print(textwrap.fill(current_state.description))
            for ob in list(current_state.objects.values()):
                print("You notice " + ob.shortdesc + ".")
            print("Health: " + str(self.health))
            print("Time elapsed: " + str(int(self.time)) + " seconds")
            print("Times moved: " + str(self.counter))
            command = input("Command> ")
            if command == "become a dog": #Easter egg
                print(" ")
                print("You are now a dog! YOU WIN THE GAME!")
                self.health = 0
            if command == "backpack": #Look in your backpack/inventory
                print(" ")
                print("You are carrying:")
                for obj in list(backpack.objects.values()):
                    print(textwrap.fill("    " + obj.shortdesc + ", which has a weight of " + str(obj.weight) + "."))
                print("You are carrying " + str(backpack.value) + " total.")
                print("You can carry up to " + str(backpack.maxweight) + " total.")
            if command == "health": #Check your health
                print(" ")
                print("You have " + str(self.health) + " health points remaining.")
            if command == "time": #Check the time elapsed
                secondspassed = int(time.time() - self.initial_time)
                minutes = secondspassed/60
                seconds = int((minutes - int(minutes)) * 60)
                print(" ")
                print(str(int(minutes)) + " minutes and " + str(seconds) + " seconds have passed since you started playing.")
            if command == "help":
                print("Your options are:")
                for e in list(current_state.edges.values()):
                    print(e.commands)
            for edge in list(current_state.edges.values()): #Use an edge
                if command in edge.commands:
                    if len(edge.to_node) > 1: #If an edge goes to more than one place, pick one at random
                        end = len(edge.to_node)
                        rand = random.randint(0, end)
                        to_n = edge.to_node[rand]
                    else:
                        to_n = edge.to_node[0]
                    confilled = 0 #Tally of how many of the constraints on an edge are filled
                    execlist = [] #List of all parts to be interpreted of the "exec" constraints on an edge
                    objlist = [] #List of all the objects that must be used to pass through an edge
                    constlist = list(edge.constraint.values())
                    for con in constlist:
                        if con.objectname == "exec":
                            confilled += 1
                            toexec = con.state.strip().split(", ")
                            for e in toexec:
                                execlist.append(e)
                        if con.state == "backpack" and con.objectname in list(backpack.objects.keys()):
                            confilled += 1
                            objlist.append(con.objectname)
                        if con.objectname == "eval" and eval(con.state) == True:
                            confilled += 1
                    if (confilled == len(constlist)) or (len(constlist) == 0): #If there are no constraints or all the constraints are filled
                        print(" ")
                        for o in objlist:
                            print("You use the " + o + ".")
                        print(textwrap.fill(edge.desc))
                        self.health = self.health + eval(edge.health)
                        self.counter += 1
                        current_state = self.states[to_n]
                        for i in execlist:
                            exec(i)      
                        break
                    else:
                        print(" ")
                        print(textwrap.fill("You cannot do this right now. Some things you might try are exploring more, waiting a bit, or picking up an object to help you."))
                    break
            for obj in list(current_state.objects.values()): #Node objects
                l = command
                if l[:4] == "get ": #Pick up an object
                    if l[4:] in obj.name:
                        if obj.weight + backpack.value <= backpack.maxweight:
                            current_state.objects.pop(obj.name[0])
                            backpack.objects[obj.name[0]] = obj
                            obj.location = "backpack"
                            backpack.value = backpack.value + obj.weight
                            print(" ")
                            print(textwrap.fill("You pick up the " + l[4:] + " and carry it with you. It has a weight of " + str(obj.weight) + "."))
                            break
                        else:
                            print(" ")
                            print(textwrap.fill("Carrying this item will put you over your weight limit. Drop another item if you want to pick up this one."))
                            break
                elif l[:4] == "look": #Look around the node for objects
                    print(" ")
                    #if len(l) <= 4:                                        #Uncomment this section and comment out the part that automatically
                    #    for obj in list(current_state.objects.values()):   #prints the objects in a node if you want the player to have to
                    #        print("There is " + obj.shortdesc + ".")       #search for objects.
                    #    break
                    if l[5:] in obj.name: #Examine an object in the room in more detail
                        print(textwrap.fill(obj.description))
                        print("It weighs " + str(obj.weight) + ".")
                        break
            for obj in list(backpack.objects.values()): #Backpack objects
                l = command
                if l[:5] == "look " and len(l) > 5:
                    if l[5:] in obj.name: #Examine an object in the backpack in more detail
                        print(" ")
                        print(textwrap.fill(obj.description))
                        print("It weighs " + str(obj.weight) + ".")
                        break
                if l[:5] == "drop ": #Drop an object in the backpack
                    if l[5:] in obj.name:
                        current_state.objects[obj.name[0]] = obj
                        backpack.objects.pop(obj.name[0])
                        obj.location = current_state.name
                        backpack.value = backpack.value - obj.weight
                        print(" ")
                        print("You drop the " + l[5:] + ".")
                        break
        print("Your health is 0. Game over.")
                        
                        
        

game = Engine("Pokemon.txt")
game.load()
game.play()
